--------------------------------------------------------------------------------
-- SubModule GPSReceiver
-- Created   12/19/2007 6:58:33 PM
--------------------------------------------------------------------------------
library IEEE;
use IEEE.Std_Logic_1164.all;
use IEEE.Numeric_Std.all;

entity GPSReceiver is port
   (
     CLK        : in    std_logic;
     RST        : in    std_logic;
     ONEPPS     : in    std_logic;
     PRESENT    : in    std_logic;
     INIT       : out   std_logic;

     RDY        : in    std_logic;
     ERR        : in    std_logic;
     D          : in    std_logic_vector(7 downto 0);

     LOAD       : out   std_logic;
     FIRST      : out   std_logic;
     DATE       : out   std_logic_vector(7 downto 0);
     OK         : out   std_logic;
     CRCERR     : out   std_logic;
     STAT       : out   std_logic_vector(255 downto 0);
     LOCK       : out   std_logic
   );
end GPSReceiver;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
architecture Structure of GPSReceiver is

-- Component Declarations

-- Constant Declarations
   constant ret1 : string := "Hb";

-- Type Declaration
   type FSM_state is (IDLE, HEADER, MSG1, TAIL);

-- Signal Declarations
   signal oldonepps               : std_logic;
   signal timeout                 : unsigned (26 downto 0); -- 2 seconds
   signal expired                 : std_logic;
   signal expire_pps              : std_logic;
   signal state                   : FSM_state;
   signal crc                     : std_logic_vector (7 downto 0);
   signal dold                    : std_logic_vector (7 downto 0);
   signal cnt                     : unsigned (7 downto 0);
   signal year                    : unsigned (15 downto 0);
   signal info                    : std_logic_vector(255 downto 0);
   signal status_locked           : std_logic;

-- Functions
   function check_string (s : in string; v : in std_logic_vector (15 downto 0))
                          return boolean is
      variable i : integer;
      --variable c : character;
      variable w : std_logic_vector (7 downto 0);
      variable ret : boolean;
   begin
      ret := true;
      for i in 1 to 2 loop
         --c := s (i);
         w := std_logic_vector (To_unsigned (character'pos(s(i)), 8));
         if w /= v (i*8-1 downto (i-1)*8) then
            ret := false;
         end if;
      end loop;
      return ret;
   end function check_string;

begin

   -- check if we received a message in the past 4 seconds
   oldonepps <= ONEPPS when rising_edge (CLK);
   process (RST, PRESENT, CLK) is
   begin
      if (RST = '1') or (PRESENT = '0') then
         expired <= '0';
         timeout <= (others => '1');
      elsif rising_edge (CLK) then
         if RDY = '1' then
            timeout <= (others => '1');
            expired <= '0';
         elsif timeout = To_unsigned (0, 27) then
            expired <= '1';
            -- status_locked <= '0';
         else
            timeout <= timeout - 1;
         end if;
      end if;
   end process;

   expire_pps <= expired and ONEPPS and not oldonepps;
   INIT <= expire_pps;
   -- INIT <= expired and not oldonepps;

   -- receive the message
   process (RST, CLK) is
   begin
      if (RST = '1') then
         state <= IDLE;
         LOAD <= '0';
         FIRST <= '0';
         DATE <= (others => '0');
         OK <= '0';
         CRCERR <= '0';
         crc <= (others => '0');
         info <= (others => '0');
         -- LOCK <= '0';
         status_locked <= '0';
      elsif rising_edge (CLK) then
         LOAD <= '0';
         FIRST <= '0';
         OK <= '0';
         CRCERR <= '0';
         case state is
            when IDLE =>
               if RDY = '1' then
                  dold <= d;
                  if check_string ("@@", d & dold) then
                     state <= HEADER;
                     cnt <= X"01";
                     crc <= (others => '0');
                  end if;
               elsif expire_pps = '1' then
                  -- LOCK <= '0';
                  status_locked <= '0';
               end if;
            -- message header
            when HEADER =>
               if RDY = '1' then
                  dold <= d;
                  crc <= crc xor d;
                  if cnt = X"00" then
                     if check_string (ret1, d & dold) then
                        state <= MSG1;
                        cnt <= To_unsigned (47, 8);
                     else
                        state <= IDLE;
                     end if;
                  else
                     cnt <= cnt - 1;
                  end if;
               end if;
            -- message
            when MSG1 =>
               if RDY = '1' then
                  dold <= d;
                  crc <= crc xor d;
                  if cnt /= X"00" then
                     cnt <= cnt - 1;
                  end if;
                  if cnt = To_unsigned (47, 8) then
                     -- month
                     DATE <= d;
                     FIRST <= '1';
                     LOAD <= '1';
                  elsif cnt = To_unsigned (46, 8) then
                     -- day
                     DATE <= d;
                     LOAD <= '1';
                  elsif cnt = To_unsigned (44, 8) then
                     -- year
                     DATE <= std_logic_vector (year (7 downto 0));
                     if year (15 downto 8) = X"00" then
                        LOAD <= '1';
                     else
                        state <= IDLE;
                     end if;
                  elsif cnt = To_unsigned (43, 8) then
                     -- hour
                     DATE <= d;
                     LOAD <= '1';
                  elsif cnt = To_unsigned (42, 8) then
                     -- minute
                     DATE <= d;
                     LOAD <= '1';
                  elsif cnt = To_unsigned (41, 8) then
                     -- second
                     DATE <= d;
                     LOAD <= '1';
                  elsif (cnt <= To_unsigned (36, 8)) and
                        (cnt >= To_unsigned (25, 8)) then
                     -- position
                     info <= info (247 downto 0) & d;
                  elsif (cnt <= To_unsigned (20, 8)) and
                        (cnt >= To_unsigned ( 1, 8)) then
                     -- speed, heading, geometry, satellites, status, tag
                     info <= info (247 downto 0) & d;
                  elsif cnt = X"00" then
                     state <= TAIL;
                  end if;
               end if;
            -- check CRC
            when TAIL =>
               if crc = X"00" then
                  OK <= '1';
                  -- LOCK <= not info(64);
                  status_locked <= info(79) and info(78) and info(77) and not info(67);
               else
                  CRCERR <= '1';
               end if;
               state <= IDLE;
         end case;
      end if;
   end process;
   year <= unsigned (dold & d) - To_unsigned (2000, 16);
   STAT <= info;
   LOCK <= status_locked;

end Structure;
--------------------------------------------------------------------------------
