--------------------------------------------------------------------------------
-- Time Decoder: Decodes a GPS time formated as MMDDYYhhmmss into GPS seconds
-- LIGO document:
-- SubModule TimeDecoder
-- Interface:
--   CLK: clock must be at least 8 times faster than load frequency
--   RESET: synchronous reset
--   LOAD: high for a clock cycle to load a new date value
--   FIRST: high when first date byte is loaded (month)
--   DATE: date byte in order: month, day, year, hour, minute, second
--   OK: indicates that correct date bytes have been clocked in
--       this is usually due to CRC check whihch is performed after the
--       date hae been received.
--       VALID goes high if the conversion was successful and the OK is high
--       OK can be high as early as last LOAD but should be before the
--       next FIRST
--   VALID: true for one cycle, if valid GPS time
--   ERR: true if error in decoding
-- Three additional clock cycles are required (with load low) to compute the
-- GPS output.
-- Created   9/25/2006 3:49:25 PM
--------------------------------------------------------------------------------
library IEEE;
use IEEE.Std_Logic_1164.all;
use IEEE.Numeric_Std.all;
--synopsys translate_off
library UNISIM;
use unisim.vcomponents.all;
--synopsys translate_on

entity TimeDecoder is port
   (
     CLK         : in    std_logic;
     RESET       : in    std_logic;
     LOAD        : in    std_logic;
     FIRST       : in    std_logic;
     DATE        : in    std_logic_vector(7 downto 0);
     OK          : in    std_logic;

     GPS         : out   std_logic_vector(31 downto 0);
     VALID       : out   std_logic;
     ERR         : out   std_logic
   );
end TimeDecoder;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
architecture Structure of TimeDecoder is

-- Components
   component ROM_1024_32
       port (
       clka:  IN  std_logic;
       addra: IN  std_logic_vector (9 downto 0);
       douta: OUT std_logic_vector (31 downto 0));
   end component;

-- Const Declarations
   constant minus_one   : signed (17 downto 0) := (others => '1');
   constant sec_in_day  : signed (17 downto 0) := to_signed (86400, 18);
   constant sec_in_hour : signed (17 downto 0) := to_signed ( 3600, 18);
   constant sec_in_min  : signed (17 downto 0) := to_signed (   60, 18);
   constant sec_in_sec  : signed (17 downto 0) := to_signed (    1, 18);

-- Type Declarations
   type decoder_type is (IDLE, MONTH, DAY, YEAR, HOUR, MINUTE, SECOND,
                         TRAIL, WAITING);

-- Signal Declarations
   signal state : decoder_type := IDLE;
   signal sec : unsigned (31 downto 0);
   signal datebyte : signed (7 downto 0);
   signal sec_valid : std_logic;
   signal decode_err : std_logic;
   signal addr : std_logic_vector (9 downto 0);
   signal lookup : std_logic_vector (31 downto 0);
   signal m1 : signed (17 downto 0);
   signal m2 : signed (17 downto 0);
   signal mul : signed (35 downto 0);
   signal oksave : std_logic;

begin
   -- output values
   GPS <= std_logic_vector (sec);
   VALID <= sec_valid;
   ERR <= decode_err;
   datebyte <= signed (DATE);
   -- multiplier
   m1 (17 downto 8) <= (others => '0');
   mul <= m1 * m2;  -- force the use of a 18x18 signed multiplier
   -- Lookup ROM
   lookup_ROM : ROM_1024_32
        port map (
            clka => CLK,
            addra => addr,
            douta => lookup);

   -- date decoder state machine
   decoder: process (CLK, RESET, LOAD, FIRST, state, datebyte) is
   begin
      if rising_edge (clk) then
         if RESET = '1' then
            state <= IDLE;
            sec <= (others => '0');
            sec_valid <= '0';
            decode_err <= '0';
            oksave <= '0';
         elsif (FIRST = '1') and (LOAD = '1') then
            state <= MONTH;
            addr (3 downto 0) <= std_logic_vector (datebyte (3 downto 0));
            sec_valid <= '0';
            if (datebyte = 0) or (datebyte > 12) then
               decode_err <= '1';
            else
               decode_err <= '0';
            end if;
            oksave <= '0';
         else
            case state is
               when IDLE =>
                  -- do nothing
                  sec_valid <='0';
               when MONTH =>
                  if LOAD = '1' then
                     state <= DAY;
                     m1 (7 downto 0) <= datebyte + X"FF";
                     m2 <= sec_in_day;
                     sec <= (others => '0');
                     if (datebyte = X"00") or (datebyte > 31) then
                        decode_err <= '1';
                     else
                        decode_err <= decode_err;
                     end if;
                  end if;
               when DAY =>
                  if LOAD = '1' then
                     state <= YEAR;
                     addr (9 downto 4) <= std_logic_vector (datebyte (5 downto 0));
                     sec <= sec + unsigned (mul (31 downto 0)); -- add days
                     if (datebyte < 7) or (datebyte > 63) then
                        decode_err <= '1';
                     else
                        decode_err <= decode_err;
                     end if;
                  end if;
               when YEAR =>
                  if LOAD = '1' then
                     state <= HOUR;
                     m1 (7 downto 0) <= datebyte;
                     m2 <= sec_in_hour;
                     if datebyte > 23 then
                        decode_err <= '1';
                     else
                        decode_err <= decode_err;
                     end if;
                  end if;
               when HOUR =>
                  if LOAD = '1' then
                     state <= MINUTE;
                     m1 (7 downto 0) <= datebyte;
                     m2 <= sec_in_min;
                     sec <= sec + unsigned (mul (31 downto 0)); -- add hours
                     if datebyte > 59 then
                        decode_err <= '1';
                     else
                        decode_err <= decode_err;
                     end if;
                  end if;
               when MINUTE =>
                  if LOAD = '1' then
                     state <= SECOND;
                     m1 (7 downto 0) <= datebyte;
                     m2 <= sec_in_sec;
                     sec <= sec + unsigned (mul (31 downto 0)); -- add minutes
                     if datebyte > 59 then
                        decode_err <= '1';
                     else
                        decode_err <= decode_err;
                     end if;
                     oksave <= OK;
                  end if;
               when SECOND =>
                  state <= TRAIL;
                  sec <= sec + unsigned (mul (31 downto 0)); -- add seconds
                  if OK = '1' then
                     oksave <= OK;
                  end if;
               when TRAIL =>
                  state <= WAITING;
                  sec <= sec + unsigned (lookup); -- add seconds of year/month
                  if OK = '1' then
                     oksave <= OK;
                  end if;
               when WAITING =>
                  if (oksave = '1') or (OK = '1') then
                     state <= IDLE;
                     sec_valid <= '1';
                  end if;
            end case;
         end if;
      end if;
   end process decoder;


end architecture Structure;



















